
from .classes import ports, marnet, passages
from .utils import get_duration, distance_length, from_nodes_edges_set, normalize_linestring, validate_lon_lat
from geojson import Feature, LineString

from .data.ports_dict import edge_list as port_e, node_list as port_n
from .data.marnet_dict import edge_list as marnet_e, node_list as marnet_n


P = from_nodes_edges_set(ports.Ports(), port_n, port_e)
M = from_nodes_edges_set(marnet.Marnet(), marnet_n, marnet_e)

del port_e, port_n, marnet_e, marnet_n


def searoute(origin, destination, units='km', speed_knot=24, append_orig_dest=False, restrictions=[passages.Passage.northwest], include_ports=False, port_params={}, M:marnet.Marnet=M, P:ports.Ports=P):
    """
    Calculates the shortest sea route between two points on Earth.

    Parameters
    ----------
    origin : a point as array lon, lat format ex. [0.35156, 50.06419]
    destination : a point as array lon, lat format ex. [117.42187, 39.36827]
    units : default is `km` = kilometers,
        can be `m` = meters `mi` = miles `ft` = foot 
        `in` = inches `deg` = degrees `cen` = centimeters 
        `rad` = radians `naut` = nautical `yd` = yards
    speed_knot : speed of the boat, default is `24` knots 
    append_orig_dest : add origin and dest geo-points in the LineString of the route, default is False
    restrictions : an list of restrictions of paths to avoid, use as per your specific need; default restricted ['northwest']; possible passages: babalmandab, bosporus, gibraltar, suez, panama, ormuz, northwest
    include_ports : boolean to include closest port close to origin or destinations
    port_params : object ; 
        - only_terminals: boolean, default False
        - country_pol: country iso code for port of load
        - country_pod: country iso code for port of discharge
        - country_restricted :  boolean, default False ; if True it will consider `country_pod` as a parameter to match with `to_cty` in ports list
    
    Returns
    -------
    a Feature (geojson) of a LineString of sea route with parameters : `unit` and `length`, `duration_hours` or port details
    """

    # Validate origin input
    validate_lon_lat(origin)
    # Validate destination input
    validate_lon_lat(destination)


    if P is None:
        raise Exception('Ports network must not be None')

    if M is None:
        raise Exception('Marnet network must not be None')

    o_origin = tuple(origin)
    o_destination = tuple(destination)

    # updating restrictions
    if restrictions is not None:
        M.restrictions = restrictions

    # H = nx.subgraph_view(G, filter_edge=filter_edge)

    if include_ports:
        if not port_params:
            port_params = {}

        only_terminals = port_params.get('only_terminals', False)
        country_pol = port_params.get('country_pol', None)
        country_pod = port_params.get('country_pod', None)
        country_restricted = port_params.get('country_restricted', False)
        country_restricted_key =  'to_cty'
        to_cty = country_pod if country_restricted else None
       

        # set origin as closest port
        closestPortOrigin = P.query(
            terminals=only_terminals, cty=country_pol, to_cty=to_cty).kdtree.query(origin)
        if closestPortOrigin:
            origin = closestPortOrigin
            port_origin = P.nodes[origin]
            if country_restricted_key in port_origin:
                del port_origin[country_restricted_key]

        # set destination as closest port
        closestPortDest = P.query(
            terminals=only_terminals, cty=country_pod).kdtree.query(destination)
        if closestPortDest:
            destination = closestPortDest
            port_dest = P.nodes[destination]
            if country_restricted_key in port_dest:
                del port_dest[country_restricted_key]
    
    # Get shortest route from the Marnet network 
    # if origin or destination is not present in M, searches from the closest one
    shortest_route_by_distance = M.shortest_path(o_origin, o_destination)

    if shortest_route_by_distance is None:
        shortest_route_by_distance = []
        
    if include_ports and shortest_route_by_distance:
        shortest_route_by_distance.insert(0, origin )
        shortest_route_by_distance.append(destination )

    if append_orig_dest:
        if (origin != o_origin):
            shortest_route_by_distance.insert(0, o_origin)
        if (destination != o_destination):
            shortest_route_by_distance.append(o_destination)

    ls = []
    previous = None

    for i in shortest_route_by_distance:
        now = i
        fixed_coords = normalize_linestring(previous, now)
        ls.append(fixed_coords)
        previous = fixed_coords

    total_length = distance_length(ls, units=units)

    duration = get_duration(speed_knot, total_length, units)

    feature = Feature(geometry=LineString(ls), properties={
                      'length': total_length, 'units': units, 'duration_hours': duration})

    if include_ports and port_origin and port_dest:
        feature.properties['port_origin'] = port_origin
        feature.properties['port_dest'] = port_dest

    return feature



class KDNode:
    """
    A KDNode
    """
    def __init__(self, point, left=None, right=None):
        self.point = point
        self.left = left
        self.right = right


class KDTree:
    """
    A KDTree


    """
    def __init__(self, points=None):
        self.k = 2
        self.root = None
        if points:
            self.root = self._build_tree(points)

    def add_point(self, point):
        if self.root is None:
            # self.k = len(point)
            self.root = KDNode(point)
        else:
            self._add_point(point, self.root, depth=0)

    def _add_point(self, point, node, depth):
        axis = depth % self.k

        if point[axis] < node.point[axis]:
            if node.left is None:
                node.left = KDNode(point)
            else:
                self._add_point(point, node.left, depth + 1)
        else:
            if node.right is None:
                node.right = KDNode(point)
            else:
                self._add_point(point, node.right, depth + 1)

    def _build_tree(self, points, depth=0):
        if not points:
            return None

        axis = depth % self.k
        sorted_points = sorted(points, key=lambda x: x[axis])
        median = len(sorted_points) // 2

        return KDNode(
            sorted_points[median],
            self._build_tree(sorted_points[:median], depth + 1),
            self._build_tree(sorted_points[median + 1:], depth + 1),
        )

    def _query(self, point, node, depth=0, best=None):
        if node is None:
            return best

        current_best = best
        current_node = node

        axis = depth % self.k
        next_best = None
        next_branch = None

        if point[axis] < node.point[axis]:
            next_node = node.left
            opposite_node = node.right
        else:
            next_node = node.right
            opposite_node = node.left

        next_best = self._query(point, next_node, depth + 1, current_best)

        if self._distance(point, current_node.point) < self._distance(point, next_best):
            next_best = current_node.point

        if abs(point[axis] - node.point[axis]) < self._distance(point, next_best):
            opposite_best = self._query(
                point, opposite_node, depth + 1, next_best)
            if self._distance(point, opposite_best) < self._distance(point, next_best):
                next_best = opposite_best

        return next_best

    def query(self, point):
        try:
            best = self.root.point
        except:
            raise Exception('Ports/Marnet network was not inititated, initiate using searoute.utils.from_nodes_edges_set function')
        return self._query(point, self.root, 0, best)

    def _distance(self, p1, p2):
        return sum((x - y) ** 2 for x, y in zip(p1, p2)) ** 0.5



import networkx as nx

from ..utils import load_from_geojson
from .kdtree import KDTree


class Ports(nx.Graph):
    """
    Base class for Ports network is an undirected graph. 

    A Ports graph stores nodes and edges with optional data, or attributes.

    Nodes are identified by an id of tuple representing a spacial location
    with (lon, lat) and it's attributes if applied.
    They form ports located around the world.


    """

    def __init__(self):
        super().__init__()
        DEFAULT_CRF = 'EPSG:3857'
        self.graph['crs'] = DEFAULT_CRF  # CRS attribute for the graph
        self.kdtree = KDTree()

    def add_node(self, node, **attr):
        if not isinstance(node, tuple):
            raise TypeError(
                "Node must be a str representing coordinates of a port.")
        x, y = node
        attr['x'] = x
        attr['y'] = y
        if not (attr['port'] and attr['cty']):
            raise TypeError(
                "Node port requires to have both port name (name), and country (cty) in properties to be correctly mapped")

        self.kdtree.add_point(node)
        super().add_node(node, **attr)

    def subgraph(self, nodes):

        subg = super().subgraph(nodes)
        subg.kdtree = KDTree(nodes)

        return subg

    def query(self, terminals: bool = True, cty: str = None, to_cty: str = None):
        """
        Query the Port graph

        Parameters
        ----------
                
        terminal: boolean, default True
            filters out only terminal marked ports
        cty : str, default None
            filters out ports in given country code (must be an ISO country code with 2 letters uppercase)
        to_cty: str, default None
            filters out ports that have to_country the port of discharge (country_pod)
            must be an ISO country code with 2 letters uppercase

        Returns
        -------
        A subgraph of Ports filtered
        """

        if not terminals and not cty:
            return self

        def cty_filter(data, cty):
            if not cty:
                return True
            else:
                return data.get('port')[:2] == cty
            
        def cty_to_filter(data, to_cty):
            if not to_cty:
                return True
            else:
                return to_cty in (data.get('to_cty') or [])

        terminal_filter = True if terminals else None

        filtered_nodes = [n for n, data in self.nodes(data=True) if data.get('t') == terminal_filter
            and cty_filter(data, cty)
            and cty_to_filter(data, to_cty)
            ]
        
        if not filtered_nodes or len(filtered_nodes)==0:
            # filter without cty as cty seems to not have ports :/
            filtered_nodes = [n for n, data in self.nodes(data=True) if data.get(
            't') == terminal_filter]
        subgraph = self.subgraph(filtered_nodes)
        return subgraph

    def filter_only_terminals(self, u, v, edge_data):
        return True if edge_data.get('t', 0) == 1 else False

    def load_geojson(self, path):
        return load_from_geojson(self, path)

    @staticmethod
    def from_geojson(path):
        return Ports().load_geojson(path)
    

    def add_edges_from_list(self, edge_list):
        if not edge_list:
            return
        
        for edge in edge_list:
            u,v,args = edge
            self.add_edge(u, v, **args)

    def add_nodes_from_list(self, node_list):
        if not node_list:
            return

        for node in node_list:
            n, args = node
            self.add_node(n, **args)

    def update_kdtree(self, nodes = None):
        if nodes:
            self.kdtree = KDTree(nodes)
        else:
            self.kdtree = KDTree(self._node)



import networkx as nx
from .passages import Passage
from ..utils import load_from_geojson, distance
from .kdtree import KDTree


class Marnet(nx.Graph):
    """Base class for maritime network is an undirected graph. 

    A Marnet stores nodes and edges with optional data, or attributes.

    Nodes are identified by an id of tuple representing a spacial location
    with (lon, lat) and it's attributes if applied.
    They form maritime routes on water.

    
    """

    def __init__(self):
        super().__init__()
        DEFAULT_CRF = 'EPSG:3857'
        self.graph['crs'] = DEFAULT_CRF  # CRS attribute for the graph
        self.restrictions = [Passage.northwest]
        self.kdtree = KDTree()

    def add_node(self, node, **attr):
        if not isinstance(node, tuple):
            raise TypeError(
                "Node must be a tuple representing the coordinates.")
        x, y = node
        attr['x'] = x
        attr['y'] = y

        self.kdtree.add_point(node)
        super().add_node(node, **attr)

    def add_edge(self, u, v, **attr):
        if not isinstance(u, tuple) or not isinstance(v, tuple):
            raise TypeError(
                "Nodes must be tuples representing the coordinates.")

        # Create nodes if they don't exist in the graph
        if u not in self:
            self.add_node(u)
        if v not in self:
            self.add_node(v)

        if not "weight" in attr:
            length = distance(u, v)
            attr["weight"] = round(length, 1)
        super().add_edge(u, v, **attr)

    def add_edges_from_list(self, edge_list):
        if not edge_list:
            return
        
        for edge in edge_list:
            u,v,args = edge
            self.add_edge(u, v, **args)

    def add_nodes_from_list(self, node_list):
        if not node_list:
            return
        
        for node in node_list:
            n, args = node
            self.add_node(n, **args)


    def subgraph(self, nodes):

        subg = super().subgraph(nodes)
        subg.kdtree = KDTree(nodes)

        return subg

    def edge_subgraph(self, edges):

        subg = super().edge_subgraph(edges)
        # subg.kdtree = KDTree([item for tpl in tuples_list for item in tpl])

        return subg

    def query(self, apply_restrictions=True, restrictions=None):
        """
        Query the Marnet graph

        Parameters
        ----------
        apply_restrictions : boolean, default True
            Filters out edges with passages out of restriction list
        restrictions : list of passages to be restricted
            A list of str, by default is None which means the restrictions of the Marnet

        Returns
        -------
        A subgraph of Marnet filtered
        """

        if apply_restrictions:
            restrictions_ = restrictions or self.restrictions
            filtered_edges = [(u, v) for u, v, data in self.edges(
                data=True) if data.get('passage', None) not in restrictions_]
            return self.subgraph(filtered_edges)
        else:
            return self

    def custom_weight(self, u, v, edge_data):
        # Function to customize edge weight based on edge attributes or conditions
        # For this example, we will avoid edges with a certain attribute (e.g., "avoid": True)
        edge_values = edge_data.values()
        avoid_edge = any(value.get('passage', None) in (
            self.restrictions) for value in edge_values)
        wights = [value.get('weight', 1.0) for value in edge_values]
        return float('inf') if avoid_edge else min(wights)

    def filter_avoid_passages(self, u, v, edge_data, args):
        edge_values = edge_data.values()

        avoid_edge = any(value.get('passage', None) in (
            self.restrictions or []) for value in edge_values)

        return not (avoid_edge)

    def load_geojson(self, *path):
        return load_from_geojson(self, *path)
    

    def update_kdtree(self, nodes = None):
        if nodes:
            self.kdtree = KDTree(nodes)
        else:
            self.kdtree = KDTree(self._node)

    # Get the shortest route by distance
    def __custom_w(self, u, v, data):
        return data.get('weight') if data.get('passage') not in self.restrictions else float('inf')

    def shortest_path(self, origin, destination):
        """
        Shortest Path between the origin and the destination.
        Dijkstra algorithm is used to perform the calculation.

        Parameters
        ----------
        origin : origin location, in the graph or not,
            if origin is not a known node, a closed node search will be performed
        destination : destination location in the graph or not
            if destination is not a known node, a closed node search will be performed

        Returns
        -------
        A list of nodes building the shortest path
        
        """
        origin_node = self.kdtree.query(origin)
        destination_node = self.kdtree.query(destination)

        return nx.shortest_path(
            self, origin_node, destination_node, weight=self.__custom_w) 

    @staticmethod
    def from_geojson(*path):
        return Marnet().load_geojson(*path)



import searoute as sr

#Define origin and destination points:
origin = [21.1545, 55.6526]
destination = [-118.2629, 33.7276]


route = sr.searoute(origin, destination, append_orig_dest=True, restrictions=['northwest'], include_ports=True, port_params={'only_terminals':True, 'country_pol': '', 'country_pod' :'', 'country_restricted': False})
# > Returns a GeoJSON LineString Feature
# show route distance with unit
print("{:.1f} {}".format(route.properties['length'], route.properties['units']))

print(route)

# Optionally, define the units for the length calculation included in the properties object.
# Defaults to km, can be can be 'm' = meters 'mi = miles 'ft' = feets 'in' = inches 'deg' = degrees
# 'cen' = centimeters 'rad' = radians 'naut' = nauticals 'yd' = yards
routeMiles = sr.searoute(origin, destination, units="mi")


from math import atan2, cos,  pow, radians, sin, sqrt, tan
import geojson
import inspect


def get_unique_number(lon, lat):
    """Get a unique hash

    Parameters
    ----------
    lon : longitude
    lat : latitude

    Returns
    -------
    A unique str of hash
    
    """
    try:
        lat_double = None
        lon_double = None
        if isinstance(lat, str):
            lat_double = float(lat)
        else:
            lat_double = lat
        if isinstance(lon, str):
            lon_double = float(lon)
        else:
            lon_double = lon

        lat_int = int((lat_double * 1e7))
        lon_int = int((lon_double * 1e7))
        val = abs((lat_int << 16 & 0xffff0000) | (lon_int & 0x0000ffff))
        val = val % 2147483647
        return val
    except Exception as e:
        print("marking OD_LOC_ID as -1 getting exception inside get_unique_number function")
        print("Exception while generating od loc id")
        return None


def speed_coef(unit):

    if unit == 'km':
        return 1.852
    elif unit == 'm':
        return 1852
    elif unit == 'mi':
        return 1.15078
    elif unit == 'ft':
        return 6076.12
    elif unit == 'in':
        return 72913, 4
    elif unit == 'deg':
        return 1.852
    elif unit == 'cen':
        return 185200
    elif unit == 'rad':
        return 1.852
    elif unit == 'yd':
        return 2025.37
    return 1


avg_earth_radius_km = 6371008.8
conversions = {
    "km": 0.001,
    "m": 1.0,
    "mi": 0.000621371192,
    "ft": 3.28084,
    "in": 39.370,
    "deg": 1 / 111325,
    "cen": 100,
    "rad": 1 / avg_earth_radius_km,
    "naut": 0.000539956803,
    "yd": 0.914411119,
    "nm": 0.00071506154
}


def distance(coordinates1, coordinates2, units: str = "km"):
    """
    Distance calculation

    Parameters
    ----------
    coordinates1 : touble of lat,lon, from location
    coordinates2 : touble of lat,lon, to location
    units : a unit, default is `km`


    Returns
    -------
    Distance in `units`


    """

    dlat = radians((coordinates2[1] - coordinates1[1]))
    dlon = radians((coordinates2[0] - coordinates1[0]))

    lat1 = radians(coordinates1[1])
    lat2 = radians(coordinates2[1])

    a = pow(sin(dlat / 2), 2) + pow(sin(dlon / 2), 2) * cos(lat1) * cos(lat2)
    b = 2 * atan2(sqrt(a), sqrt(1 - a))
    return b * avg_earth_radius_km * conversions[units]


def distance_length(line: [tuple], units: str = "km"):
    """
    Length of a line of coordinates 

    Parameters
    ---------
    line : a list of tuple [(lon, lat), (lon, lat), ...]
    units: the unit, default is `km`

    Returns
    -------
    Distance in `units`
    """
    if line is None:
        return 0

    length = 0
    coords = line
    le = len(coords)-1
    for i, point in enumerate(coords):
        if i != le:
            length += distance(point, coords[i+1], units)

    return length


def get_duration(speed_knot, length, units):
    """Duration 


    Parameters
    ----------
    speed_knot : seed of the boat in knot
    length : length or distance in `units`
    units : the unit of the length/distance

    Returns
    -------
    The duration in hours `hr`

    """
    duration = 0

    if (speed_knot > 0):
        speed_knot = speed_knot*speed_coef(units)
        duration = length/speed_knot

    return duration


def find_lowest_key(data, *keys):
    min_dict_ids = tuple(
        min(data, key=lambda dict_id: data[dict_id].get(key, float('inf')))
        for key in keys
    )

    return min_dict_ids


def __default_filter(u, *something):
    """
    A default filter is a lazy filter and will return always True

    Parameters
    ----------
    u : a node, not used
    something : arguments of something, not used

    Returns
    --------
    Always `True`


    """
    return True


def nearest_node(G, *args, filter=None, fargs=()):
    """
    Nearest node in the graph

    Returns
    -------
    filter can be a True, callable function which should return a function
    when filter = True it means validated

    """
    dists = {}
    keys = set(range(len(args)))
    found = False
    ignoreEdgeCheck = False

    if filter is None:
        filter = __default_filter

    f_params = len(inspect.signature(filter).parameters)

    if not f_params in [2, 4]:
        raise Exception('filter should have 2 or 4 parameters')

    if f_params == 2:
        ignoreEdgeCheck = True

    for node, data in G.nodes(data=True):
        x = data.get('x', None)
        y = data.get('y', None)

        filtered = False
        if ignoreEdgeCheck:
            filtered = filter(node, fargs)

        for ix, arg in enumerate(args):
            arg = tuple(arg)
            aDist = distance((x, y), arg)
            if node not in dists:
                dists[node] = {}

            edge_data = {}
            if not ignoreEdgeCheck and G.has_edge(node, arg):
                edge_data = G.get_edge_data(node, arg)

            if filtered or (not ignoreEdgeCheck and filter(node, arg, edge_data, fargs)):
                dists[node][ix] = aDist
                found = True

    if not found:
        # return nothing if not found anything
        return None

    return find_lowest_key(dists, *keys)


def to_nodes_edges_list(G, file_name):
    """
    Converts a graph to nodes, edges list of dict, and exports to a file

    Parameters
    ----------
    G : a graph of Marnet or Ports
    file_name : the name of the file to be saved

    Returns
    -------
    Void
    
    """
    if not file_name:
        file_name = 'graph_data.py'

    with open(file_name, 'w') as convert_file:
        convert_file.write(f"node_list={str(list(G.nodes(data=True)))}\n\nedge_list={str(list(G.edges(data=True)))}")

def to_nodes_edges_set(G, file_name):
    if not file_name:
        file_name = 'graph_data.py'

    with open(file_name, 'w') as convert_file:
        convert_file.write(f"node_list={str(G._node)}\n\nedge_list={str(G._adj)}")

def from_nodes_edges_set(G, node_set, edge_set):
    """Returns a searoute Network (Ports or Marnet) from a set of nodes and edges.

    Parameters
    ----------
    G : a Ports or Marnet network (instance)
        Ports or Marnet object are searoute classes inheriting the NetworkX graph.


    node_set : dictorary of nodes, required.
        A set of nodes represnation with its attribites.
        A node should be a spacial location as a tuple of `lon`, `lat` (lon, lat)
        Must contain at least 2 attributes : `x` = lon, and `y` = lat

    edge_set : dictorary of edges, required.
       A set of edges represnation with its edge-data.

    Examples
    --------
    >>> my_nodes = {(1,2):{'x':1, 'y':2}, (2,2):{'x':2, 'y':2}}  # nodes of representing (1,2) of lon = 1, lat = 2
    >>> my_edges = {(1,2):{(2,2):{weight:10, other_attr:'some_value'}}} # (1,2) -> (2,2) with weight representing the distance or additional attributes
    >>> # recognized attributes are : `weight` (distance), ``passagepassage`` (name of the `passage` to be restricted by M.restrictions) 
    >>> M = sr.from_nodes_edges_set( sr.Marnet() ,my_nodes, my_edges)

    or

    >>> P = sr.from_nodes_edges_set( sr.Ports() ,my_nodes, None) # for Ports

    """


    if G is None:
        raise Exception('G should not be none or provide create_with')
    
    G._node = node_set or {}
    G._adj = edge_set or {}
    G.update_kdtree(node_set)

    return G

def from_nodes_edges_list(G, node_list, edge_list):
    #if type(G) in [Marnet, Ports] :
    G.add_nodes_from_list(node_list)
    G.add_edges_from_list(edge_list)

    return G

    #else:
    #    raise Exception(f"{type(G)} not supported")

def load_from_geojson(G, *geojson_file):
    for gf in geojson_file:
        with open(gf, 'r') as f:
            data = geojson.load(f)

        def handle_geometry(geometry, properties):
            if geometry.type == 'LineString':
                coords = geometry.coordinates
                for u, v in zip(coords[:-1], coords[1:]):
                    G.add_edge(tuple(u), tuple(v), **properties)
                    # the other side as well
                    G.add_edge(tuple(v), tuple(u), **properties)
            elif geometry.type == 'MultiLineString':
                for line_string in geometry.coordinates:
                    for u, v in zip(line_string[:-1], line_string[1:]):
                        G.add_edge(tuple(u), tuple(v), **properties)
                        # the other side as well
                        G.add_edge(tuple(v), tuple(u), **properties)
            elif geometry.type == 'Point':
                coords = tuple(geometry.coordinates)
                G.add_node(coords, **properties)
            elif geometry.type == 'MultiPoint':
                for point_coords in geometry.coordinates:
                    coords = tuple(point_coords)
                    G.add_node(coords, **properties)
            else:
                # Handle other geometries if needed (e.g., MultiPolygon, Polygon, etc.)
                pass

        # Extract CRS information from FeatureCollection
        if data.type == 'FeatureCollection':
            G.graph['crs'] = data.get('crs', {}).get(
                'properties', {}).get('name', None) or G.graph['crs']

            for feature in data.features:
                handle_geometry(feature.geometry, feature.properties)
        else:
            handle_geometry(data.geometry, data.properties)

    return G


def normalize_linestring(prev, now):
    """
    Normalize the coordinate for the lineString.
    The previus (`prev`) parameter will dictate the direction of the string

    Parameters
    ----------
    prev : a coord of (lon, lat)
    now : a coord to fix of format (lon, lat)

    Returns
    -------
    a cordinate normalized or fixt coordinate of format (lon, lat)

    """
    if not prev:
        return now
    
    now_x, now_y = now
    prev_x, prev_y = prev

    if not prev_x is None:
        px = prev_x-now_x
        if px < -180:
            now_x = now_x - 360
        elif px > 180:
            now_x = now_x + 360
    
    return (now_x, now_y)


def validate_lon_lat(coord):
    """
    Validate a (lon, lat) coordinate.

    Parameters
    ----------
    coord (tuple or list): A tuple or list containing two elements, representing longitude and latitude.

    Raises
    ------
    ValueError: If the input format is incorrect or if the longitude and latitude values are invalid.

    Returns
    -------
    None: The function does not return anything. It either validates successfully or raises an exception.
    """

    # Check if the input is a tuple or list with two elements
    if not isinstance(coord, (tuple, list)) or len(coord) != 2:
        raise ValueError("Invalid input format. Must be a tuple or list with two elements as (lan,lon).")

    lon, lat = coord

    # Latitude should be between -90 and 90 degrees
    # Longitude can be any value
    if not (-90 <= lat <= 90) or not isinstance(lon, (int, float)):
        raise ValueError("Invalid longitude and/or latitude.")
    

def normalize_longitude(longitude:float):
    """
    Normalize a longitude value to the range of -180 to 180 degrees.

    Parameters:
    - longitude (float): The input longitude value.

    Returns:
    - float: The normalized longitude value within the range of -180 to 180 degrees.

    Examples:
    normalize_longitude(200)
    >>> -160.0

    normalize_longitude(-270)
    >>> 90.0
    """
    return (longitude % 360 + 540) % 360 - 180